% echo "compile([benchmark]). allgoals." | sicstus

%    positive 0 -> false
%    positive (s _) -> true
%    add 0 X -> X
%    add (s X) Y -> s (add X Y)
%    half(0) -> 0
%    half(s 0) -> 0
%    half(s(s X)) -> s(half(X))
%    coin -> 0
%    coin -> s 0
%    big -> s 0
%    big -> s (s 0)
%
%    coinaf -> alt(0, s(0))
%    bigaf -> s(alt(0, s(0)))
%    alt X Y -> X
%    alt X Y -> Y
%
%    coinac -> or(0, s(0))
%    bigac -> s(or(0, s(0)))

% Auxiliary predicate to built natural numbers 
% from usual notation to Peano notation.
built_nat(0, 0) :- !.
built_nat(I, s(N)) :- NextI is (I - 1), built_nat(NextI, N).

%%%%%%%%%
% Evaluation to normal form (nf).
nf(T, T) :- var(T), !.
nf(T, NfT) :- hnf(T, HnfT), nf1(HnfT, NfT). 
           
nf1(0, 0).
nf1(s(X), s(NfX)) :- nf(X,NfX).
nf1(true,true).
nf1(false,false).
nf1(or(X,Y), or(NfX,NfY)) :- nf(X,NfX), nf(Y,NfY).

%%%%%%%%%
% Evaluation to head normal form (hnf).
hnf(T,T) :- var(T), !.
hnf(positive(X), H) :- !, positive(X, H).
hnf(add(X1,X2), H) :- !, add(X1,X2,H).
hnf(half(X), H) :- !, half(X, H).
hnf(coin, H):- !, coin(H).
hnf(big, H):- !, big(H).
hnf(coinaf, H):- !, coinaf(H).
hnf(bigaf, H):- !, bigaf(H).
hnf(alt(X1,X2), H) :- !, alt(X1,X2,H).
hnf(coinac, H):- !, coinac(H).
hnf(bigac, H):- !, bigac(H).

hnf(T,T). % the term T is in hnf; a constant is in hnf

%%%%%%%%%
% Clauses to implement the "positive" defined fuction
positive(X, H) :- hnf(X, HX), positive_1(HX, H).
positive_1(0, false).
positive_1(s(_), true).
positive_1(or(X,Y), H) :- hnf(or(positive(X),positive(Y)), H).


%%%%%%%%%
% Clauses to implement the "add" defined fuction
add(X1,X2,H) :- hnf(X1, HX1), add_1(HX1, X2, H).
add_1(0, X2, H) :- hnf(X2, H).
add_1(s(X3), X2, H) :- hnf(s(add(X3,X2)), H).
add_1(or(X,Y), Z, H) :- hnf(or(add(X,Z),add(Y,Z)), H).

%%%%%%%%%
% Clauses to implement the "half" defined fuction
half(X, H) :- hnf(X, HX), half_1(HX, H).
half_1(0, 0).
half_1(s(X), H) :- hnf(X,HX), half_2(HX, H).
half_1(or(X,Y), H) :- hnf(or(half(X),half(Y)), H).
half_2(0, 0).
half_2(s(X), H) :- hnf(s(half(X)), H).
half_2(or(X,Y), H) :- hnf(or(half(s(X)),half(s(Y))), H).

% original half_1(0, 0).
% original half_1(s(X), H) :- hnf(X,HX), half_2(s(HX), H).
% original half_1(or(X,Y), H) :- hnf(or(half(X),half(Y)), H).
% original half_2(s(0), 0).
% original half_2(s(s(X)), H) :- hnf(s(half(X)), H).
% original half_2(s(or(X,Y)), H) :- hnf(or(half(s(X)),half(s(Y))), H).

% wrong half_1(0, 0).
% wrong half_1(s(0), 0).
% wrong half_1(s(s(X)), H) :- !, hnf(s(half(X)), H).
% wrong half_1(or(X,Y), H) :- hnf(or(half(X),half(Y)), H).
% wrong half_1(s(or(X,Y)), H) :- !, hnf(or(half(s(X)),half(s(Y))), H).
% wrong half_1(s(X), H) :- hnf(X,HX), half_1(s(HX), H).

%%%%%%%%%
% Clauses to implement the "coin" defined fuction
coin(H) :- hnf(0, H).
coin(H) :- hnf(s(0), H).

% Clauses to implement the "big" defined fuction
big(H) :- hnf(s(0), H).
big(H) :- hnf(s(s(0)), H).

%%%%%%%%%
% Clauses to implement the "coinaf" defined fuction
coinaf(H) :- hnf(alt(0, s(0)), H).

% Clauses to implement the "bigaf" defined fuction
bigaf(H) :- hnf(s(alt(0, s(0))), H).

%%%%%%%%%
% Clauses to implement the "alt" defined function
alt(X1,_,H) :- hnf(X1,H).
alt(_,X2,H) :- hnf(X2,H).

%%%%%%%%%
% Clauses to implement the "coinac" defined fuction
coinac(H) :- hnf(or(0, s(0)), H).

% Clauses to implement the "bigac" defined fuction
bigac(H) :- hnf(s(or(0, s(0))), H).

% This is how time and space are measured.
stat_start(Bytes) :-
      garbage_collect, nogc,
      statistics(global_stack,[Bytes,_]).
stat_end(Bytes) :- statistics(global_stack,[Bytes,_]). 

stat_all(S,X) :-
  write(S), write(X), nl,
  % ---- THE NUMBER 40 IS SYSTEM DEPENDENT ----
  stat_start(U), nf(X, R), stat_end(V), W is (V-U-40)//4,
  write('Mem: '), write(W), write('  '),
  write('result: '), write(R), nl, fail.
stat_all(_,_).

g0   :- stat_all('goal 0: ',_).
g1   :- stat_all('goal 1: ',positive(add(big, 0))).
g2   :- stat_all('goal 2: ',add(s(0),coin)).
g3   :- built_nat(100, N), stat_all('goal 3: ',add(N, coin)).
g4   :- stat_all('goal 4: ',half(big)).
gaf1 :- stat_all('goal AF 1: ',positive(add(bigaf, 0))).
gaf2 :- stat_all('goal AF 2: ',add(s(0),coinaf)).
gaf3 :- built_nat(100, N), stat_all('goal AF 3: ',add(N, coinaf)).
gaf4 :- stat_all('goal AF 4: ',half(bigaf)).
gac1 :- stat_all('goal AC 1: ',positive(add(bigac, 0))).
gac2 :- stat_all('goal AC 2: ',add(s(0),coinac)).
gac3 :- built_nat(100, N), stat_all('goal AC 3: ',add(N, coinac)).
gac4 :- stat_all('goal AC 4: ',half(bigac)).

allgoals :- 
   g0, g1, gaf1, gac1,
   g0, g2, gaf2, gac2,
   g0, g3, gaf3, gac3,
   g0, g4, gaf4, gac4.



